import {
  format, ucFirst, truncate, trimAll, replaceWords,
  indexOf, lastIndexOf, insert, append, prefix,
  getStringValue, quotGetters, surroundWith,
  toCamelcase, wordsFirstUp, toDashedNotation, toSnakeCase,
} from "./instanceMethods.js";

import {xString as $S, isNumber} from "./genericMethods.js";
export default instanceCreator;

function instanceCreator({initialstring, customMethods} = {}) {
  let customStringExtensions = { };
  let instance = new Proxy(customStringExtensions, getTraps(customStringExtensions));
  let actualValue = initialstring?.constructor === String ? initialstring : ``;
  let history = [actualValue];
  const enumerable = false, configurable = false;
  
  Object.defineProperties(customStringExtensions, { 
    append: { value(...strings) { return wrap(append(actualValue, ...strings)); }, enumerable, configurable },
    enclose: { value(start, end) { return wrap(surroundWith(actualValue, start, end)); }, enumerable, configurable },
    format: { value(...tokens) { return wrap(format(actualValue, ...tokens)); }, enumerable, configurable },
    indexOf: { value(str) { return indexOf(actualValue, str); }, enumerable, configurable },
    interpolate: { value(...tokens) { return wrap(format(actualValue, ...tokens)); }, enumerable, configurable },
    insert: { value({ value, values, at } = {}) { 
        return wrap(insert(actualValue, { value, values, at })); 
      }, enumerable 
    },
    lastIndexOf: { value(str) { return lastIndexOf(actualValue, str); }, enumerable, configurable },
    prefix: { value(...strings) { return wrap(prefix(actualValue, ...strings)); }, enumerable, configurable },
    replaceWords: { value({caseSensitive = false, replacements = {}} = {}) {
      return wrap(replaceWords(actualValue, {replacements, caseSensitive}));
    }, enumerable, configurable },
    toString: { value() { return actualValue; }, enumerable, configurable },
    truncate: { value({at, html = false, wordBoundary = false} = {}) {
      return wrap(truncate(actualValue, {at, html, wordBoundary})); }, enumerable, configurable },
    valueOf: { value() { return actualValue; }, enumerable, configurable },
    undoLast: { value(nSteps) { return undoSteps(nSteps); }, enumerable, configurable },

    camelCase: { get() { return wrap(toCamelcase(getStringValue(actualValue))); }, enumerable, configurable },    
    clone: { get() { return clone(); }, enumerable, configurable },
    constructor: { get() { return $S.constructor; }, enumerable, configurable },
    firstUp: { get() { return wrap(ucFirst(getStringValue(actualValue))); }, enumerable, configurable },
    history: { 
      get() { return history; },
      set(value) { history = value; },
      enumerable, configurable },
    empty: { get() { return actualValue.length < 1; } },
    notEmpty: { get() { return actualValue.length < 1 ? undefined: instance; } },
    kebabCase: { get() { return wrap(toDashedNotation(getStringValue(actualValue))); }, enumerable, configurable },
    quote: quotGetters(instance, wrap),
    snakeCase: { get() { return wrap(toSnakeCase(getStringValue(actualValue))); } },
    trimAll: { get() { return wrap(trimAll(actualValue)); }, enumerable, configurable },
    trimAllKeepLF: { get() { return wrap(trimAll(actualValue, true)); }, enumerable, configurable },
    undoAll: { get() { return undoAll(); }, enumerable, configurable },
    undo: { get() { return undoLast(); }, enumerable, configurable },
    wordsUCFirst: { get() { return wrap(wordsFirstUp(getStringValue(actualValue))); }, enumerable, configurable },
    value: { 
      get() { return actualValue; }, 
      set(value) { 
        actualValue = getStringValue(value).length ? value : actualValue;
        history.push(actualValue);
      }, enumerable, configurable },
  });
  
  injectCustomMethods(customMethods);
  
  return Object.freeze(instance);
  
  function getTraps(extensions) {
    return {
      get( target, key ) {
        return key in extensions
          ? extensions[key]
          : key in String.prototype
            ? wrapNative(key)
            : undefined;
      },
    };
  }
  
  function clone() {
    const newInstance = $S(actualValue);
    newInstance.history = [...history];
    return newInstance;
  }
  
  function wrapNative(key) {
    return actualValue[key] instanceof Function
      ? function(...args) {
        const result = actualValue[key](...args); 
        return result?.constructor === String ? wrap(actualValue[key](...args)) : result; 
      } : actualValue[key];
  }
  
  function undoAll() {
    while(history.length > 1) { history.pop(); }
    actualValue = history.at(-1);
    return wrap(actualValue, false);
  }
  
  function undoSteps(steps) {
    if (!isNumber(steps)) {
      return wrap(actualValue, false); 
    }
    
    const historyLen = history.length;
    
    if (steps >= historyLen || steps < 1) {
      history = history.slice(0, 1);
      actualValue = history.at(-1); 
      return wrap(history.at(-1), false);  
    }
    
    history = history.slice(0, historyLen - steps);
    
    actualValue = history.at(-1);
    return wrap(actualValue, false);
  }
  
  function undoLast() {
    if (history.length === 1) {
      return wrap(history[0]);
    }
    history.pop();
    actualValue = history.at(-1);
    return wrap(actualValue, false);
  }
  
  function wrap(result, pushHistory = true) {
    const changed = actualValue !== result;
    changed && pushHistory && history.push(result);
    actualValue = result;
    return instance;
  }
  
  function injectCustomMethods(customMethods) {
    Object.entries(customMethods).forEach(([methodName, methodContainer]) => {
      const {enumerable, method, isGetter} = methodContainer;
      const descriptor = isGetter
        ? { get() { return wrap(method(instance).value); }, enumerable, configurable }
        : { value(...args) { return wrap(method(instance, ...args).value); }, enumerable, };
      
      Object.defineProperty(customStringExtensions, methodName, descriptor);
    });
  }
}
